from __future__ import unicode_literals

from .service import Service
from .settings import Settings
from .connection.client import Client
from .connection.server import Server
from . import storage
from . import events


__version__ = "0.6.1"
__license__ = "BSD"
__author__ = "Richard Terry"
__credits__ = [
]



"""
Timers
"""
from __future__ import unicode_literals

class Registry(object):
    def __init__(self, service):
        self.service = service
        self._next = None
    
    def get_next(self):
        if self._next.time <= self.service.time:
            yield self._next


class Timer(object):
    def __init__(self):
        pass



"""
Test library
"""
from __future__ import print_function
from __future__ import unicode_literals

import os
import socket
import telnetlib
import time
import threading
import unittest

import mara
from mara import styles
from mara.settings import defaults


__all__ = [
    # Constants
    'DEBUG', 'DEBUG_TN', 'ATTEMPT_MAX', 'ATTEMPT_SLEEP', 'ATTEMPT_TIMEOUT',
    'IMPOSSIBLE', 'NEWLINE', 'CWD', 'EXAMPLES_DIR', 
    
    # Useful vars and classes
    'thread_counter', 'unittest', 'TestCase', 'TestService', 'Client', 'hr',
]


# Debug everything that we do
DEBUG = False

# Debug everything that telnetlib does
DEBUG_TN = False

# Limits for attempting to start the server, and to connect to the server
ATTEMPT_MAX = 10
ATTEMPT_SLEEP = 0.1

# Timeout for socket reading
ATTEMPT_TIMEOUT = 1

# A string which will not be returned from the server
IMPOSSIBLE = '$$--__--$$'
NEWLINE = '\r\n'

# Example scripts expect themselves to be in the root path
# Find the examples dir and use that as the root path
CWD = os.getcwd()
EXAMPLES_DIR = os.path.join(CWD, 'examples')
if not os.path.isdir(EXAMPLES_DIR):
    raise ValueError('Examples dir not found')

# Global thread counter to generate unique ids
class ThreadCounter(object):
    def __init__(self):
        self.count = 0
    def get(self):
        self.count += 1
        return self.count
thread_counter = ThreadCounter()


class FakeClient:
    columns = 80
    class service:
        class settings:
            hr_sequence = '-'
            hr_state = None

def hr(*msg):
    return styles.hr(*msg).render(FakeClient(), styles.StatePlain())

class TestCase(unittest.TestCase):
    def assertLine(self, response, expected):
        """
        Assert equal, but stripping the trailing newline from response
        """
        if not response.endswith(NEWLINE):
            self.fail(
                (
                    'Response does not end in newline sequence: '
                    'expected "%s", received "%s"'
                ) % (expected, response)
            )
        return self.assertEqual(response[:-len(NEWLINE)], expected)


class TestService(object):
    """
    Wrapper for a Service instance being tested
    
    Pass settings on the constructor, or set them on the settings attribute
    """
    # Global settings
    settings = mara.settings.Settings(
        root_path=EXAMPLES_DIR,
        log='all' if DEBUG else False,
        settings_collect_args=False,
    )
    
    stores = property(lambda self: self.service.stores)
    
    def __init__(self, *args, **kwargs):
        """
        Create and run a service
        """
        # Give service a thread
        self.thread_id = thread_counter.get()
        
        # Create service
        self.service = self.define()
        
        # Collect settings
        new_settings = mara.settings.Settings(*args, **kwargs)
        if self.settings:
            self.settings.update(new_settings)
        else:
            self.settings = new_settings
        
        # Start service thread
        self._exception = None
        self.thread = threading.Thread(
            name="%s-%s" % (self.__class__.__name__, self.thread_id),
            target=self.run,
        )
        self.thread.daemon = True  # Kill with main process
        self.thread.start()
        
        # Wait for service to run (or fail)
        while (
            not self._exception
            and not getattr(self.service, 'server', None)
            and not getattr(self.service.server, 'running', False)
        ):
            time.sleep(ATTEMPT_SLEEP)
        
        # Catch exception
        if self._exception:
            self.thread.join()
            raise RuntimeError("Thread %s failed" % self.thread.name)
    
    def define(self):
        """
        Define the service
        """
        raise NotImplementedError('Test service must implement define()')
        
    def run(self):
        """
        Run the service. Call this as a thread target.
        """
        try:
            self.service.run(self.settings)
        except Exception as e:
            self._exception = e
            raise
    
    def stop(self):
        """
        Stop the service
        """
        self.service.stop()
        self.thread.join()



class Client(object):
    """
    Telnet client
    
    Wrapper for telnetlib.Telnet
    """
    username_prompt = 'Username: '
    password_prompt = 'Password: '
    
    def __init__(self, tests=None, host=defaults.host, port=defaults.port):
        """
        Create a connection to the client
        """
        self.tests = tests
        self.host = host
        self.port = port
        self.open()
    
    def open(self):
        """
        Create a connection to the client
        """
        # Try ATTEMPT_MAX times, in case socket isn't quite ready yet
        n = 0
        while n < ATTEMPT_MAX:
            n += 1
            try:
                self.tn = telnetlib.Telnet(self.host, self.port)
            except socket.error:
                time.sleep(ATTEMPT_SLEEP)
                if n == ATTEMPT_MAX:
                    raise
            else:
                break
        
        # Turn on telnetlib debugging
        if DEBUG_TN:
            self.tn.set_debuglevel(1)

    
    def close(self):
        self.tn.close()
    
    def write(self, *lines):
        """
        Write one or more lines
        """
        for line in lines:
            if DEBUG:
                print("test write>:", line)
            line += NEWLINE
            self.tn.write(line.encode('utf-8'))
    
    def read(self):
        """
        Read the next line
        """
        response = self.read_until(NEWLINE, timeout=ATTEMPT_TIMEOUT)
        if DEBUG:
            print("test read>:", response[:-len(NEWLINE)])
        return response
    
    # Map fns through to tn
    def read_until(self, expected, timeout=ATTEMPT_TIMEOUT):
        response = self.tn.read_until(expected.encode('utf-8'), timeout)
        return response.decode('utf-8')
    
    def read_all(self):
        return self.tn.read_all().decode('utf-8')


    #
    # Helper functions - common tasks
    #
    
    def assertNothing(self):
        """
        Assert that there is nothing to read
        """
        if not self.tests:
            raise ValueError('Cannot have a client assert without self.tests')
        self.tests.assertFalse(self.tn.sock_avail())
    
    def assertRead(self, line):
        """
        Assert that the exact response is received
        """
        if not self.tests:
            raise ValueError('Cannot have a client assert without self.tests')
        response = self.read_until(line, timeout=ATTEMPT_TIMEOUT)
        if DEBUG:
            print("test read> :", response)
        self.tests.assertEqual(response, line)
        
    def assertLine(self, *lines):
        """
        Assert that the specified response is received
        """
        if not self.tests:
            raise ValueError('Cannot have a client assert without self.tests')
        for line in lines:
            self.tests.assertLine(self.read(), line)
    
    def assertResponse(self, send, *lines):
        """
        Assert that the data sent receives the specified response
        """
        if not self.tests:
            raise ValueError('Cannot have a client assert without self.tests')
        self.write(send)
        self.assertLine(*lines)
    
    def assertClosed(self):
        if not self.tests:
            raise ValueError('Cannot have a client assert without self.tests')
        
        with self.tests.assertRaises(EOFError) as cm:
            self.read_until(IMPOSSIBLE, timeout=ATTEMPT_TIMEOUT)
        self.tests.assertEqual(str(cm.exception), 'telnet connection closed')
    
    def login(self, username, password=None):
        self.read_until(self.username_prompt)
        self.write(username)
        if password:
            self.read_until(self.password_prompt)
            self.write(password)



"""
Test the example mud
"""
from __future__ import unicode_literals

from .lib import *
import mara
import os
import shutil

from examples import mud
from examples.mud.rooms import room_lobby

MUD_STORE = os.path.join(EXAMPLES_DIR, 'test_mud_store')


class MudTestService(TestService):
    settings = TestService.settings + mara.settings.Settings(
        store_path=MUD_STORE,
    )
    def define(self):
        return mud.service

class MudClient(Client):
    username_prompt = 'What is your name? '
    password_prompt = 'Enter your password, or press enter to pick a new name: '
    
    def login(self, *args, **kwargs):
        super(MudClient, self).login(*args, **kwargs)
        self.assertLine('')
    
    def create_account(self, username, password):
        self.read_until(self.username_prompt)
        self.write(username)
        self.read_until('Do you want to create an account? (Enter yes or no) ')
        self.write('yes')
        self.read_until('Enter a password: ')
        self.assertResponse(password, '')
        self.assertRead('Confirm password: ')
        self.assertResponse(password, '', 'Account created!')
        
    def quit(self):
        self.assertResponse(
            'quit',
            hr('Goodbye!'),
        )
        self.assertClosed()


class MudTest(TestCase):
    """
    Test the example mud
    """
    def remove_store(self):
        # Sanity check store so we don't delete the wrong thing
        if not MUD_STORE or not CWD or not MUD_STORE.startswith(CWD):
            raise ValueError('Invalid talker store')
        
        # Ensure test store doesn't exist
        if os.path.isdir(MUD_STORE):
            shutil.rmtree(MUD_STORE)
        
    def setUp(self):
        # Start service
        self.remove_store()
        self.service = MudTestService()
        
    def tearDown(self):
        self.service.stop()
        self.remove_store()
    
    def create_accounts(self):
        """
        Create accounts for 3 users
        """
        ann = MudClient(self)
        ann.create_account('ann', 'password1')
        ann.assertLine(
            room_lobby.name,
            room_lobby.desc[0],
            'Nobody else is here.',
            'There are exits to the north, to the south and to the east.',
        )
        
        bob = MudClient(self)
        bob.create_account('bob', 'password2')
        bob.assertLine(
            room_lobby.name,
            room_lobby.desc[0],
            'ann is here.',
            'There are exits to the north, to the south and to the east.',
        )
        ann.assertLine(
            '-- bob has connected --',
            'bob appears from nowhere'
        )
        
        cat = MudClient(self)
        cat.create_account('cat', 'password3')
        cat.assertLine(
            room_lobby.name,
            room_lobby.desc[0],
            'ann and bob are here.',
            'There are exits to the north, to the south and to the east.',
        )
        ann.assertLine(
            '-- cat has connected --',
            'cat appears from nowhere',
        )
        bob.assertLine(
            '-- cat has connected --',
            'cat appears from nowhere',
        )
        
        return ann, bob, cat
    
    
    def test_rooms(self):
        """
        Test rooms, exits, movement and clone rooms
        """
        # Collect yaml rooms
        room_pool = self.service.stores['room'].manager.get('pool')
        clone_zone = self.service.stores['room'].manager.get('clone_zone')
        
        # Create accounts
        ann, bob, cat = self.create_accounts()
        
        # Check commands work to all in room
        ann.assertResponse('say Hello', 'You say: Hello')
        bob.assertLine('ann says: Hello')
        cat.assertLine('ann says: Hello')
        
        ann.assertResponse('dance', 'You dance')
        bob.assertLine('ann dances')
        cat.assertLine('ann dances')
        
        # Fake exit
        ann.assertResponse(
            'south',
            "You think about leaving, but decide it's nice here.",
        )
        bob.assertNothing()
        cat.assertNothing()
        
        # Exit north
        ann.assertResponse(
            'north',
            'You go north',
            room_pool.name,
            room_pool.desc[0],
            'Nobody else is here.',
            'There is one exit to the south.',
        )
        bob.assertLine('ann leaves to the north')
        cat.assertLine('ann leaves to the north')
        
        # Commands only work in the room
        ann.assertResponse('say Hello', 'You say: Hello')
        bob.assertNothing()
        cat.assertNothing()
        
        # Socials only work in the room
        ann.assertResponse('dance', 'You dance')
        bob.assertNothing()
        cat.assertNothing()
        
        # Return south
        ann.assertResponse(
            'south',
            'You go south',
            room_lobby.name,
            room_lobby.desc[0],
            'bob and cat are here.',
            'There are exits to the north, to the south and to the east.',
        )
        bob.assertLine('ann enters from the north')
        cat.assertLine('ann enters from the north')

        # Ann enters the clone zone
        ann.assertResponse(
            'east',
            'You go east',
            clone_zone.name,
            clone_zone.intro[0],
            clone_zone.desc[0],
            'Nobody else is here.',
            'There is one exit to the west.',
        )
        bob.assertLine('ann leaves to the east')
        cat.assertLine('ann leaves to the east')
        
        # Bob enters the clone zone
        bob.assertResponse(
            'east',
            'You go east',
            clone_zone.name,
            clone_zone.intro[0],
            clone_zone.desc[0],
            'Nobody else is here.',
            'There is one exit to the west.',
        )
        ann.assertNothing()
        cat.assertLine('bob leaves to the east')

        # Ann and bob are alone in the clone zone
        ann.assertResponse(
            'look',
            clone_zone.name,
            # Look ma, no intro
            clone_zone.desc[0],
            'Nobody else is here.',
            'There is one exit to the west.',
        )
        ann.assertResponse('say Hello', 'You say: Hello')
        bob.assertNothing()
        cat.assertNothing()
        
        # Return from the cold
        ann.assertResponse(
            'west',
            'You go west',
            room_lobby.name,
            room_lobby.desc[0],
            'cat is here.',
            'There are exits to the north, to the south and to the east.',
        )
        bob.assertNothing()
        cat.assertLine('ann enters from the east')




# -*- coding: utf-8 -*-
"""
Test the example chat server
"""
from __future__ import unicode_literals

from .lib import *
from examples import chat


class ChatTestService(TestService):
    def define(self):
        return chat.service

class ChatClient(Client):
    username_prompt = 'Welcome. What is your name? '

class ChatTest(TestCase):
    """
    Test the example chat server
    
    Multiple tests of same service to check service.stop() cleans up correctly.
    """
    def setUp(self):
        self.service = ChatTestService()
        
    def tearDown(self):
        self.service.stop()
    
    def test_single(self):
        "Test chat single connection"
        client = ChatClient(self)
        client.login('testuser')
        client.assertLine('Welcome, testuser')
        client.assertResponse('Hello', 'testuser> Hello')
        client.assertResponse('/me dances', 'testuser dances')
        client.assertResponse('/who', 'Users here at the moment:', '  testuser')
        client.assertResponse('/quit', 'Goodbye!')
        client.assertClosed()

    def test_multiple(self):
        "Test chat multiple connections"
        client1 = ChatClient(self)
        client1.login('ann')
        client1.assertLine('Welcome, ann')
        
        client2 = ChatClient(self)
        client2.login('bob')
        client2.assertLine('Welcome, bob')
        client1.assertLine('-- bob has connected --')
        
        client1.assertResponse('Hello', 'ann> Hello')
        client2.assertLine('ann> Hello')
        
        client1.assertResponse('/me dances', 'ann dances')
        client2.assertLine('ann dances')
        
        client1.assertResponse(
            '/who',
            'Users here at the moment:', '  ann', '  bob',
        )
        client2.assertNothing()
        
        client1.assertResponse('/quit', 'Goodbye!')
        client1.assertClosed()
        client2.assertLine('-- ann has disconnected --')
        
        client2.assertResponse('/quit', 'Goodbye!')
        client2.assertClosed()

    def test_unicode(self):
        "Test chat unicode"
        client = ChatClient(self)
        client.login('testuser')
        client.assertLine('Welcome, testuser')
        client.assertResponse('cash $£€ monies', 'testuser> cash $£€ monies')
        client.assertResponse(
            'niño, 男の子, เด็กผู้ชาย, garçon',
            'testuser> niño, 男の子, เด็กผู้ชาย, garçon'
        )
        client.assertResponse('/quit', 'Goodbye!')
        client.assertClosed()



"""
Test the example echo server
"""
from __future__ import unicode_literals

from .lib import *
from examples import echo


class EchoTestService(TestService):
    def define(self):
        return echo.service
    
class EchoTest(TestCase):
    """
    Test the example echo server
    
    Multiple tests of same service to check service.stop() cleans up correctly.
    """
    def setUp(self):
        self.service = EchoTestService()
        
    def tearDown(self):
        self.service.stop()
    
    def test_single(self):
        "Test echo single connection"
        client = Client(self)
        client.assertResponse('Hello', 'Hello')
        client.close()

    def test_multiple(self):
        "Test echo multiple connections"
        client1 = Client(self)
        client2 = Client(self)
        client1.write('client1')
        client2.write('client2')
        client1.assertLine('client1')
        client2.assertLine('client2')
        client1.close()
        client2.close()



"""
Mara process manager

Controls settings, server, and loaded modules
"""
from __future__ import unicode_literals

from collections import defaultdict
import datetime
import inspect
import six
import sys
import time

from .angel import Process
from .container import ClientContainer
from .connection import Server
from .connection.client import client_registry
from .settings import collect as collect_settings
from . import logger
from . import events
from . import timers


class Service(ClientContainer):
    """
    Service management
    """
    def __init__(self):
        super(Service, self).__init__()
        
        # Store settings
        self.settings = None
        
        # Angel currently unknown
        self.angel = None
        
        # No server yet
        self.server = None
        
        # Store of stores
        self.stores = defaultdict(dict)
        
        # Consistent time for events and timers
        self.time = time.time()
        
        # Initialise events
        self.events = defaultdict(list)
        self._known_events = {}
        
        # Initialise timers
        self.timers = timers.Registry(self)
        
        # An empty log
        self.log = None
    
    datetime = property(
        fget = lambda self: datetime.datetime.fromtimestamp(self.time),
        doc = "Get the current time as a datetime object"
    )
    
    
    #
    # Events
    #
    
    def listen(self, event_class, handler=None):
        """
        Bind a handler to the specified event class, and to its subclasses
        """
        # Called directly
        if handler:
            if isinstance(handler, events.handler.HandlerType):
                handler = handler()
            self._listen(event_class, handler)
            return

        # Called as a decorator
        def decorator(fn):
            if isinstance(fn, events.handler.HandlerType):
                fn = fn()
            self._listen(event_class, fn)
            return fn
        return decorator
        
    def _listen(self, event_class, handler):
        """
        Internal method to recursively bind a handler to the specified event
        class and its subclasses. Call listen() instead.
        """
        # Recurse subclasses. Do it before registering for this event in case
        # they're not known yet, then they'll copy handlers for this event
        for subclass in event_class.__subclasses__():
            self._listen(subclass, handler)
            
        # Register class
        self._ensure_known_event(event_class)
        self.events[event_class].append(handler)
    
    def _ensure_known_event(self, event_class):
        """
        Ensure the event class is known to the service.
        
        If it is not, inherit handlers from its first base class
        """
        # If known, nothing to do
        if event_class in self._known_events:
            return
        
        # If base class isn't an event, nothing to do
        base_cls = event_class.__bases__[0]
        if not isinstance(base_cls, events.Event):
            return
        
        # Ensure base class is known, and copy its handlers
        self._ensure_known_event(base_cls)
        self.events[event_class] = self.events[base_cls][:]
    
    def trigger(self, event):
        """
        Trigger the specified event
        """
        # Make sure we've seen this event class before
        self._ensure_known_event(event.__class__)
        
        # Make sure all listeners have access to the service, in case they're
        # defined out of scope
        event.service = self
        
        self.log.event(event)
        for handler in self.events[event.__class__]:
            # Catch stopped event
            if event.stopped:
                return
                
            # Run handler co-routine or function
            if (
                inspect.isgeneratorfunction(handler)
                or isinstance(handler, events.Handler)
            ):
                generator = handler(event)
                try:
                    next(generator)
                except StopIteration:
                    pass
                else:
                    if hasattr(event, 'client'):
                        event.client.capture(generator)
            else:
                handler(event)
    
    #
    # Timers
    #
    
    def timer(self, cls=timers.PeriodTimer, **kwargs):
        """
        Function decorator to define a timer
        """
        def wrap(fn):
            timer = cls(**kwargs)
            timer.fn = fn
            self.timers.add(timer)
            return fn
        return wrap

        
    #
    # Container overrides
    #
    
    # Clients attribute returns server's clients with service global filter
    clients = property(
        lambda self: self.filter_all(self, self.server._clients.values())
    )
    def add_client(self, client):
        raise NotImplementedError('Cannot add a client to a service')
    def remove_client(self, client):
        raise NotImplementedError('Cannot remove a client from a service')
    
    @staticmethod
    def _filter_all_default(service, clients, **kwargs):
        return clients
    _filter_all = _filter_all_default
    
    @property
    def filter_all(self):
        return self._filter_all
    
    @filter_all.setter
    def filter_all(self, value):
        if value is None:
            value = self._filter_all_default
        elif not callable(value):
            raise ValueError('Filter must be a callable')
        self._filter_all = value


    #
    # Internal operations
    #
    
    def run(self, *args, **kwargs):
        """
        Collect settings and start the server
        """
        # Collect settings
        self.settings = collect_settings(*args, **kwargs)
        
        # Try to connect to angel, and initialise appropriate logger
        self.angel = Process(self)
        if self.angel.connect():
            self.log = logger.AngelLogger(self.settings, self.angel)
            self.log.service('Connected to angel')
        else:
            self.log = logger.Logger(self.settings)
            # Not running with angel - remove from log levels
            self.log.ignore_angel()
            self.log.update_prefix()
            self.log.service('No angel detected')
        
        # Inform that we're starting
        self.log.service('Service starting')
        self.trigger(events.PreStart())
        
        # Try to deserialise service from angel
        if self.angel:
            # Connected to angel
            serialised = self.angel.get_service()
            if serialised:
                self.deserialise(serialised)
                if self.angel:
                    self.log.service('Service restarted')
                    self.trigger(events.PostStart())
                    self.trigger(events.PostRestart())
            
            if not self.angel:
                self.log.service('Angel lost - aborting startup')
                return
            
            # Set up angel watcher - when the angel dies, we die too
            @self.timer(period=1)
            def angel_watcher(timer):
                self.angel.poll()
                if not self.angel:
                    self.log.service('Angel lost - terminating')
                    self.write_all('-- Angel lost --')
                    self.stop()
            
        # Start new server if no angel, or angel didn't have serialised service
        if not self.server:
            self.server = Server(self)
            self.trigger(events.PostStart())
        
        # If we have an angel, tell it we have started successfully
        if self.angel:
            self.angel.started()
            
        self.server.listen()
        
    def poll(self):
        """
        A poll from the server
        """
        # Update time
        self.time = time.time()
        
        # Call due and overdue timers
        self.timers.trigger()
    
    def stop(self):
        """
        Stop the server
        """
        self.trigger(events.PreStop())
        self.server.shutdown()
        self.log.service('Service stopped')
        self.trigger(events.PostStop())

    def restart(self):
        """
        Restart the process
        """
        if not self.angel:
            raise ValueError('Cannot restart, not connected to angel')
        
        # Send reset warning
        self.log.service('Service restarting')
        self.trigger(events.PreRestart())
        
        # Flush all server's client output buffers
        # (bypass our .clients attribute, that's filtered)
        for client in self.server._clients.values():
            client.flush()
        
        # Suspend server, so sockets will back up waiting for new process
        self.server.suspend()
        
        # Serialise
        serialised = self.serialise()
        
        # Send serialised service to angel and tell it we're stopping
        self.angel.set_service(serialised)
        
        # Terminate immediately
        self.log.service('Service terminating')
        sys.exit()

    def serialise(self):
        """
        Serialise to be passed to a new process
        """
        # Serialise the store of stores (with session data)
        store_data = defaultdict(dict)
        for store_name, store in self.stores.items():
            store_data[store_name] = store.manager.serialise(session=True)
        
        return {
            'server':    self.server.serialise(),
            'store_data': store_data,
        }

    def deserialise(self, serialised):
        """
        Deserialise a serialised service into this instance
        """
        # Restore server and clients (using their old ids)
        self.server = Server(self, serialised=serialised['server'])
        
        # Deserialise store of stores (with session data)
        store_data = serialised['store_data']
        for store_name, data in store_data.items():
            store_cls = self.stores.get(store_name)
            if store_cls is None:
                self.log.store(
                    'Could not thaw store "%s" - no longer defined' % store_name
                )
                continue
            store_cls.manager.deserialise(data, session=True)
        
        # Update all known clients' ids
        clients = client_registry.values()
        client_registry.clear()
        for client in clients:
            client.update_id()



"""
Logging class for Mara
"""
from __future__ import print_function
from __future__ import unicode_literals

import os
import sys
import datetime


# Log verbosity constants
NONE        = ''
ALL         = 'all'
ANGEL       = 'angel'
SERVICE     = 'service'
SERVER      = 'server'
CLIENT      = 'client'
EVENT       = 'event'
STORE       = 'store'
DEBUG       = 'debug'
DEFAULT_LEVELS = ['angel', 'server']

class Logger(object):
    def __init__(self, settings):
        """
        Determine which log levels to listen to, and open log file
        """
        # Get settings
        self.settings = settings
        levels = settings.log
        self.filename = settings.get_path('log_file')
        self.with_pid = settings.log_pid
        self.with_time = settings.log_time
        self.with_level = settings.log_level
        
        # Start disabled
        self.file = None
        self.prefix = ''
        self.disabled = True
        
        # Default levels
        if levels is True:
            levels = DEFAULT_LEVELS
        
        # If no levels, logging is disabled
        if not levels:
            self.levels = []
            self.exclude = []
            return
        
        # See what to listen to and what to exclude
        if hasattr(levels, 'split'):
            levels = levels.split(',')
        self.levels = [level for level in levels if not level.startswith('-')]
        self.exclude = [level[1:] for level in levels if level.startswith('-')]
        
        if ALL in self.levels:
            self.levels = ALL
        if ALL in self.exclude:
            raise ValueError('Log level -all is not valid')
            
        # Now open logger
        self.disabled = False
        self.open()
        
    def open(self):
        """
        Open or re-open the log file
        
        Called automatically by constructor
        
        Call this if:
            * the verbosity level changes from or to NONE
            * the process ID changes
        """
        # Close the logfile, if logger is enabled
        self.close()
        
        # See if this logger is supposed to be turned off, or
        # if we're printing to STDOUT
        if self.disabled or not self.filename:
            return
        
        # Try to open file
        try:
            self.file = open(self.filename, 'a')
        except IOError as e:
            print("Error opening logfile: %s" % e, file=sys.stderr)
            self.disabled = True
        
        # Pre-calculate prefix
        self.update_prefix()
    
    def ignore_angel(self):
        """
        Ignore the angel level
        """
        self.exclude.append(ANGEL)
    
    def force_with_pid(self):
        """
        Set with_pid to True, unless with_pid is False
        """
        if self.with_pid is None:
            self.with_pid = True
            self.update_prefix()

    def update_prefix(self):
        """
        Pre-calculate prefix settings
        """
        # Re-calculate with_level==None - levels may have changed
        if self.settings.log_level is None:
            if len(self.levels) == 1:
                self.with_level = False
            else:
                self.with_level = True
        
        # Generate as much of the prefix as possible
        self.prefix = ''
        if self.with_pid or (self.filename and self.with_pid is None):
            self.prefix = '[%s] ' % os.getpid()
        elif not self.filename and not self.with_level:
            # If printing to stdout, start each line with a *
            self.prefix = '* '
        
    def write(self, level, *lines):
        """
        Filter by level before writing lines to the log file
        """
        # Skip if disabled, or unwanted log level
        if self.disabled or level in self.exclude or (
            self.levels != ALL and level not in self.levels
        ):
            return
        
        # Add pid prefix
        prefix = self.prefix
        if self.with_time or (self.filename and self.with_time is None):
            # Format as close to syslog as possible
            prefix = datetime.datetime.today().strftime('%b %d %X') + " " + prefix
        if self.with_level:
            prefix = '%s%s> ' % (prefix, level)
        lines = ['%s%s' % (prefix, line) for line in lines]
        
        self._write(lines)
    
    def _write(self, lines):
        """
        Perform actual write, without filtering or processing
        """
        # See if printing to STDOUT
        if not self.file:
            print("\n".join(lines))
            return
        
        # Write lines
        self.file.write("\n".join(lines) + "\n")
        self.file.flush()
    
    def close(self):
        """
        Close the log file
        """
        if self.file:
            self.file.close()
    
    def __del__(self):
        """
        Clean up
        """
        # The log file should get closed by file.__del__, but just in case
        self.close()
    
    # Write aliases
    def angel(self, *lines):
        self.write(ANGEL, *lines)
    def service(self, *lines):
        self.write(SERVICE, *lines)
    def server(self, *lines):
        self.write(SERVER, *lines)
    def client(self, *lines):
        self.write(CLIENT, *lines)
    def event(self, *lines):
        self.write(EVENT, *lines)
    def store(self, *lines):
        self.write(STORE, *lines)
    def debug(self, *lines):
        self.write(DEBUG, *lines)


class AngelLogger(Logger):
    """
    Logger which writes to the angel
    """
    def __init__(self, settings, angel):
        self._angel = angel
        super(AngelLogger, self).__init__(settings)
        
        # Angel always logs with pid
        self.force_with_pid()
        
    def open(self):
        "Nothing to open"
    
    def close(self):
        "Nothing to close"
    
    def _write(self, lines):
        self._angel.log(lines)



"""
Test the example talker
"""
from __future__ import unicode_literals

from .lib import *
import os
import shutil

import mara
from examples import talker

TALKER_STORE = os.path.join(EXAMPLES_DIR, 'test_talker_store')


class TalkerTestService(TestService):
    settings = TestService.settings + mara.settings.Settings(
        store_path=TALKER_STORE,
    )
    def define(self):
        return talker.service

class TalkerClient(Client):
    username_prompt = 'What is your name? '
    password_prompt = 'Enter your password, or press enter to pick a new name: '
    
    def login(self, *args, **kwargs):
        super(TalkerClient, self).login(*args, **kwargs)
        self.assertLine('')
    
    def create_account(self, username, password):
        self.read_until(self.username_prompt)
        self.write(username)
        self.read_until('Do you want to create an account? (Enter yes or no) ')
        self.write('yes')
        self.read_until('Enter a password: ')
        self.assertResponse(password, '')
        self.assertRead('Confirm password: ')
        self.assertResponse(password, '', 'Account created!')
        
    def quit(self):
        self.assertResponse(
            'quit',
            hr('Goodbye!'),
        )
        self.assertClosed()


class TalkerTest(TestCase):
    """
    Test the example talker
    
    Multiple tests of same service to check service.stop() cleans up correctly.
    """
    def remove_store(self):
        # Sanity check store so we don't delete the wrong thing
        if not TALKER_STORE or not CWD or not TALKER_STORE.startswith(CWD):
            raise ValueError('Invalid talker store')
        
        # Ensure test store doesn't exist
        if os.path.isdir(TALKER_STORE):
            shutil.rmtree(TALKER_STORE)
        
    def setUp(self):
        # Start service
        self.remove_store()
        self.service = TalkerTestService()
        
    def tearDown(self):
        self.service.stop()
        self.remove_store()
    
    def test_account_create(self):
        "Test manual account creation in talker"
        client = TalkerClient(self)
        client.assertLine('Welcome to the Mara example talker!', '')
        client.assertRead('What is your name? ')
        client.assertResponse('ann', 'There is nobody with that name.')
        client.assertRead('Do you want to create an account? (Enter yes or no) ')
        client.assertResponse(
            'yes',
            'Please pick a password for your account.',
            'Your password must be at least 6 characters long.',
        )
        client.assertRead('Enter a password: ')
        client.assertResponse(
            'short', '', 'Your password must be at least 6 characters long.',
        )
        client.assertRead('Enter a password: ')
        client.assertResponse('password1', '')
        client.assertRead('Confirm password: ')
        client.assertResponse(
            'password1',
            '',
            'Account created!',
            'Welcome, ann!',
            'Nobody else is here.',
        )
        client.quit()
        
        # Check we can now log in
        client.open()
        client.login('ann', 'password1')
        client.assertLine('Welcome, ann!', 'Nobody else is here.')
        client.quit()
        
    def test_multiple(self):
        "Test talker multiple connections"
        # Create accounts
        ann = TalkerClient(self)
        ann.create_account('ann', 'password1')
        ann.assertLine('Welcome, ann!', 'Nobody else is here.')
        
        bob = TalkerClient(self)
        bob.create_account('bob', 'password2')
        bob.assertLine('Welcome, bob!', 'ann is here.')
        ann.assertLine('-- bob has connected --')
        
        cat = TalkerClient(self)
        cat.create_account('cat', 'password3')
        cat.assertLine('Welcome, cat!', 'ann and bob are here.')
        ann.assertLine('-- cat has connected --')
        bob.assertLine('-- cat has connected --')
        
        # Try private commands
        ann.assertResponse('admin', 'There are no admin users')
        ann.assertResponse(
            'commands',
            hr('Commands'), (
                'admin commands emote gender help look password quit say '
                'tell users who'
            ), hr(),
        )
        ann.assertResponse(
            'help',
            'Syntax: help <command>, or type commands to see a list of commands',
        )
        ann.assertResponse(
            'help commands',
            hr('Help: commands'),
            'List commands',
            '',
            'Syntax: commands (groups|<group>)',
            hr(),
        )
        ann.assertResponse(
            'who',
            hr('Currently online'),
            'ann\t0 seconds ago',
            'bob\t0 seconds ago',
            'cat\t0 seconds ago',
            hr(),
        )
        bob.assertNothing()
        cat.assertNothing()
        
        # Try commands which put something out to all
        ann.assertResponse('say Hello', 'You say: Hello')
        bob.assertLine('ann says: Hello')
        cat.assertLine('ann says: Hello')
        
        ann.assertResponse('emote is here', 'ann is here')
        bob.assertLine('ann is here')
        cat.assertLine('ann is here')
        
        ann.assertResponse('tell bob a secret', 'You tell bob: a secret')
        bob.assertLine('ann tells you: a secret')
        cat.assertNothing()
        
        ann.assertResponse(
            'tell bob, cat a secret', 'You tell bob and cat: a secret',
        )
        bob.assertLine('ann tells you and cat: a secret')
        cat.assertLine('ann tells you and bob: a secret')
        
        ann.assertResponse(
            'look',
            'bob and cat are here.',
        )
        bob.assertLine('ann looks around')
        cat.assertLine('ann looks around')
        
        ann.assertResponse('dance', 'You dance')
        bob.assertLine('ann dances')
        cat.assertLine('ann dances')
        
        ann.assertResponse('dance bob', 'You dance with bob')
        bob.assertLine('ann dances with you')
        cat.assertLine('ann dances with bob')
        
        # Pronoun social with gender other (default)
        ann.assertResponse('gender', 'Your gender is currently other')
        ann.assertResponse(
            'dance around bob, pointing at cat and ann',
            'You dance around bob, pointing at cat and yourself',
        )
        bob.assertLine('ann dances around you, pointing at cat and themselves')
        cat.assertLine('ann dances around bob, pointing at you and themselves')
        
        # Pronoun social with gender male
        ann.assertResponse('gender male', 'Your gender is now male')
        ann.assertResponse(
            'dance around bob, pointing at cat and ann',
            'You dance around bob, pointing at cat and yourself',
        )
        bob.assertLine('ann dances around you, pointing at cat and himself')
        cat.assertLine('ann dances around bob, pointing at you and himself')
        
        # Pronoun social with gender femmale
        ann.assertResponse('gender female', 'Your gender is now female')
        ann.assertResponse(
            'dance around bob, pointing at cat and ann',
            'You dance around bob, pointing at cat and yourself',
        )
        bob.assertLine('ann dances around you, pointing at cat and herself')
        cat.assertLine('ann dances around bob, pointing at you and herself')



"""
Useful functions
"""
from __future__ import unicode_literals

import datetime

TIME_UNITS = ['day', 'hour', 'minute', 'second']


def detail_error():
    from traceback import format_exception
    import sys
    return [e.strip() for e in format_exception(*sys.exc_info())]

def pretty_list(data):
    if not data:
        return ''
    last = data[-1]
    rest = data[:-1]
    if rest and last:
        last = ' and ' + last
    return ', '.join(rest) + last

def pretty_age(seconds=None, now=None, then=None):
    if not seconds:
        seconds = now - then
    
    sec_delta = datetime.timedelta(seconds = now - then)
    age = dict(zip(
        TIME_UNITS,
        [
            sec_delta.days,
            sec_delta.seconds // 3600,
            sec_delta.seconds // 60 % 60,
            sec_delta.seconds % 60
        ]
    ))
    for attr in TIME_UNITS:
        if age[attr] > 0:
            return "%d %s%s ago" % (age[attr], attr, '' if age[attr]==1 else 's')
    return '0 seconds ago'



"""
Client-aware string styling and effects
"""
from __future__ import unicode_literals

__all__ = [
    'String',
    'normal', 'bold', 'faint', 'italic', 'underline', 'negative', 'strike',
    'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white',
    'hr',
]


# ANSI_CSI<style>[;<style>]ANSI_FINAL
ANSI_CSI = '\033['
ANSI_END = 'm'

# Style codes
NORMAL      = '0'
BOLD        = '1'
FAINT       = '2'
ITALIC      = '3'
UNDERLINE   = '4'
NEGATIVE    = '7'
STRIKE      = '9'

# Foreground colours
RED         = '31'
GREEN       = '32'
YELLOW      = '33'
BLUE        = '34'
MAGENTA     = '35'
CYAN        = '36'
WHITE       = '37'


class State(object):
    """
    Hold style state and render the ANSI sequence
    """
    # SGR codes
    codes = None
    
    def __init__(self, *codes):
        self.codes = list(codes)
    
    def append(self, code):
        self.codes.append(code)
    
    def pop(self):
        self.codes.pop()
    
    def render(self):
        """
        Render the ANSI escape sequence
        """
        return ANSI_CSI + ';'.join(
            # Always reset, in case we removed a style
            ['0']
            # Add all SGR codes (don't bother deduping)
            + self.codes
        ) + ANSI_END
    
    def __add__(self, state):
        return self.__class__(*(self.codes + state.codes))


class StatePlain(State):
    """
    Version of State which won't render any ANSI sequences
    """
    def render(self):
        return ''


class String(object):
    """
    A container of a list of strings or String instances, which concatenates
    and modifies them when rendered, based on the Client and State.
    """
    def __init__(self, *content):
        self.content = content
    
    def render(self, client, state):
        # Render content to string
        out = state.render()
        for bit in self.content:
            if isinstance(bit, String):
                bit = bit.render(client, state)
            out += bit
        return out
    
    def __add__(self, other):
        return self.__class__(*(self.content + [other]))


class Style(String):
    # The SGR code
    code = None
    
    def render(self, client, state):
        # Add code and render state
        state.append(self.code)
        
        # Add content
        out = super(Style, self).render(client, state)
        
        # Reset state
        state.pop()
        
        return out


# Styles
class normal(Style):    code = NORMAL
class bold(Style):      code = BOLD
class faint(Style):     code = FAINT
class italic(Style):    code = ITALIC
class underline(Style): code = UNDERLINE
class negative(Style):  code = NEGATIVE
class strike(Style):    code = STRIKE

# Foreground colours
class red(Style):       code = RED
class green(Style):     code = GREEN
class yellow(Style):    code = YELLOW
class blue(Style):      code = BLUE
class magenta(Style):   code = MAGENTA
class cyan(Style):      code = CYAN
class white(Style):     code = WHITE


class hr(String):
    """
    Horizontal rule, with centered content
    """
    def render(self, client, state):
        # Use settings state as default (if set)
        if client.service.settings.hr_state:
            state = client.service.settings.hr_state + state
        
        # Build a pattern the size of the client's terminal
        sequence = client.service.settings.hr_sequence
        len_sequence = len(sequence)
        pattern = (
            # Ceiling division to build it bigger and cut it down
            sequence * ((client.columns // len_sequence) + 1)
        )[:client.columns]
        
        # If no content, skip the rest and return with state
        if not self.content:
            return state.render() + pattern
        
        # Calculate left and right padding using width of content without style
        text_only = super(hr, self).render(client, StatePlain())
        space = (client.columns - (len(text_only) + 2))
        left = space // 2
        right = left
        if space % 2:
            right += 1
        
        # If it's a wide sentence or a narrow term, ensure we've got something
        if left < len_sequence:
            left = len_sequence
        if right < len_sequence:
            right = len_sequence
        
        # Render again but with style, and return with state
        out = super(hr, self).render(client, state)
        return state.render() + pattern[:left] + ' ' + out + ' ' + pattern[-right:]



"""
Angel

Wrapper to start mara process, keep it running, and manage restarts
"""
from __future__ import unicode_literals

import multiprocessing
from multiprocessing.connection import Listener, Client
import os
import select
import socket
import subprocess
import sys
import time

from . import settings
from . import logger


# Commands sent by process to angel
CMD_GET_SERVICE = 'GET_SERVICE' # New process starting; get service from angel
CMD_SET_SERVICE = 'SET_SERVICE' # Old process stopping; send service to angel
CMD_STARTED = 'STARTED'         # New process has started
CMD_POLL = 'POLL'               # Process is checking in on angel
CMD_LOG = 'LOG'                 # Log a line from the process
CMD_OK = 'OK'                   # Command received and OK

# Start delay is doubled every time it fails, until it breaks this limit
# When 30, delays will be: 0, 1, 2, 4, 8, 16, 32, 32, 32, ...
START_DELAY_MAX = 30

# Time to wait after calling angel.stop(), to allow for the socket to release
STOP_DELAY = 0.1


class Angel(object):
    """
    The angel which manages mara script processes
    """
    args = None
    settings = None
    running = False
    active_process = None
    
    def __init__(self, *args, **kwargs):
        self.args = self.collect_args()
        self.settings = settings.collect(*args, **kwargs)
        
        # Move root_path to script's path
        if self.settings.root_path is None:
            # Get the dir name of the script, and either ensure it's an
            # absolute path, or join it to the cwd
            self.settings.root_path = os.path.abspath(
                os.path.dirname(self.args[0])
            )
        
        # Start logger
        self.log = logger.Logger(self.settings)
        self.log.force_with_pid()
    
    def collect_args(self):
        """
        Collect args for starting the Mara service
        """
        sys.argv.pop(0)
        if len(sys.argv) < 1:
            raise ValueError('First argument must be path to Mara service')
        return sys.argv

    def run(self):
        """
        Main angel loop
        """
        # Open angel socket
        authkey = self.settings.angel_authkey.encode('utf-8')
        listener = Listener(
            str(self.settings.get_path('angel_socket')),
            self.settings.angel_family,
            authkey=authkey,
        )
        multiprocessing.current_process().authkey = authkey
        
        # Give the listener a fileno so we can use it directly with select
        listener.fileno = listener._listener._socket.fileno
        
        # Main angel loop
        socket_to_process = {}
        process_to_socket = {}
        old_process = None
        start_delay = 0
        self.running = True
        while self.running:
            # We don't have an active process, so there shouldn't be others
            # If there are, terminate them
            for process, process_socket in list(process_to_socket.items()):
                if process.poll():
                    process.terminate()
                process_socket.close()
                del process_to_socket[process]
                del socket_to_process[process_socket]
            
            # There is nothing valid in the store
            # Either this is the first time it's running,
            # or it failed, so didn't get a chance to serialise anything
            store = None
            
            # Start script, with increasing delay if it last start failed
            if start_delay:
                time.sleep(start_delay)
                if start_delay < START_DELAY_MAX:
                    start_delay *= 2
            else:
                start_delay = 1
            self.active_process = self.start_process()
            
            # Loop while the script is running
            while self.active_process.poll() is None:
                # Check old processes for dead
                for process in process_to_socket.keys():
                    if process == self.active_process:
                        # Already checked active process
                        continue
                    if process.poll() is not None:
                        self.log.angel('Process %s has died' % process.pid)
                        process_to_socket[process].close()
                        del process_to_socket[process]
                        del socket_to_process[process_socket]
                
                # Check sockets for activity
                read_sockets = []
                send_sockets = []
                try:
                    read_sockets, send_sockets = select.select(
                        [listener] + list(socket_to_process.keys()), [], [], 1,
                    )[0:2]
                except select.error as e:
                    self.log.angel('Angel select error: %s' % e)
                except socket.error as e:
                    self.log.angel('Angel socket error: %s' % e)
                
                # Check listener socket for new connections
                if listener in read_sockets:
                    # Only leave process sockets on read_sockets
                    del read_sockets[read_sockets.index(listener)]
                    
                    # Accept the connection
                    process_socket = listener.accept()
                    
                    # Check this makes sense
                    if self.active_process in process_to_socket:
                        # We already have a socket for the active process.
                        # This shouldn't happen - only explanation is this is
                        # an intruder. Ignore it.
                        self.log.angel(
                            'New connection for existing process %s ignored' %
                            self.active_process.pid
                        )
                        process_socket.close()
                    
                    # Store against the current process
                    self.log.angel(
                        'Established connection to process %s' %
                        self.active_process.pid
                    )
                    socket_to_process[process_socket] = self.active_process
                    process_to_socket[self.active_process] = process_socket
                    
                # Check process sockets who want to say something
                for process_socket in read_sockets:
                    process = socket_to_process[process_socket]
                    try:
                        raw = process_socket.recv()
                    except Exception as e:
                        # We were told there would be something to read
                        process_socket.close()
                        try:
                            process.terminate()
                        except OSError:
                            # Already terminated
                            pass
                        del process_to_socket[process]
                        del socket_to_process[process_socket]
                        continue
                    
                    # Get command and data
                    if len(raw) == 2:
                        cmd, data = raw
                    else:
                        self.log.angel(
                            'Invalid data received from process %s' %
                            process.pid
                        )
                        cmd, data = None, None
                    
                    # Process cmd
                    if cmd == CMD_GET_SERVICE:
                        # Process has started and wants serialised data
                        process_socket.send((CMD_OK, store))
                        
                        if store is not None:
                            self.log.angel(
                                'Sent service to process %s' % process.pid
                            )
                        
                        # Empty store - can only be deserialised once
                        store = None
                        
                    elif cmd == CMD_SET_SERVICE:
                        # Process is sending us serialised service data
                        store = data
                        
                        # Process is now waiting for the OK before it stops.
                        # Don't send it yet - only stop once we get CMD_STARTED
                        # from the new process.
                        old_process = self.active_process
                        
                        self.log.angel(
                            'Received service from process %s' % process.pid
                        )
                        
                        # Start new process
                        self.active_process = self.start_process()
                        
                    elif cmd == CMD_STARTED:
                        # New process has started; cmd sent after PostStart
                        process_socket.send((CMD_OK, None))
                        
                        # It's not going to fall over immediately, reset delay
                        start_delay = 0
                        
                        # Tell old process it's ok to die now
                        if old_process:
                            if old_process in process_to_socket:
                                process_socket = process_to_socket[old_process]
                                process_socket.send((CMD_OK, None))
                            
                            self.log.angel(
                                'Process %s active, terminating process %s' % (
                                    process.pid, old_process.pid
                                )
                            )
                        else:
                            self.log.angel('Process %s active' % process.pid)

                    elif cmd == CMD_POLL:
                        process_socket.send((CMD_OK, None))
                    
                    elif cmd == CMD_LOG:
                        process_socket.send((CMD_OK, None))
                        self.log._write(data)
                        
                    else:
                        # Something has gone wrong - kill them
                        self.log.angel(
                            'Command not recognised, terminating process %s' %
                            process.pid
                        )
                        process_socket.close()
                        if process.poll() is None:
                            process.terminate()
                        del process_to_socket[process]
                        del socket_to_process[process_socket]
                        continue
            
            # Active process has died
            if self.running:
                # It died unexpectedly
                self.log.angel(
                    'Active process %s died unexpectedly' % self.active_process.pid
                )
            else:
                # We're shutting down
                self.log.angel(
                    'Active process %s terminating' % self.active_process.pid
                )
            
            if self.active_process in process_to_socket:
                # Clean up
                process_socket = process_to_socket[self.active_process]
                process_socket.close()
                del process_to_socket[self.active_process]
                del socket_to_process[process_socket]
    
    def start_process(self):
        """
        Start the Mara service in a separate process
        """
        process = subprocess.Popen([sys.executable] + self.args)
        self.log.angel('Starting process %s' % process.pid)
        return process
    
    def stop(self):
        """
        Terminate the current active process, wait for it to close, and stop
        the main angel loop
        """
        self.running = False
        
        if self.active_process.poll() is None:
            self.active_process.terminate()
            self.active_process.wait()
            # It should be ok, but give the system a moment to free the port
            time.sleep(STOP_DELAY)


class Process(object):
    """
    A process managed by an angel
    """
    def __init__(self, service):
        self.service = service
        self.settings = service.settings
        self.client = None
    
    @property
    def _log(self):
        return self.service.log
    
    def connect(self):
        """
        Try to connect to angel
        
        Returns True if connection is made, and instance evaluates to True
        Returns False if angel is not found, and instance evaluates to False
        """
        # Cannot log - service.log is defined based on outcome of this
        try:
            authkey = self.settings.angel_authkey.encode('utf-8')
            self.client = Client(
                str(self.settings.get_path('angel_socket')),
                self.settings.angel_family,
                authkey=authkey,
            )
            multiprocessing.current_process().authkey = authkey
        except socket.error:
            return False
        return True
    
    def _send_ok(self, cmd, data=None, log=None):
        """
        Send data to the angel and wait for the OK
        
        Return any response data received with the OK
        
        If angel fails, self.client will be cleared to None, so instance will
        evaluate to False
        """
        if not self.client:
            raise ValueError('Cannot send %s, not connected to angel' % cmd)
        try:
            self.client.send((cmd, data))
            # Now block and wait for the OK - this may take a while
            # The new process has to start and take over the sockets first
            ok, response = self.client.recv()
        except Exception as e:
            if cmd != CMD_LOG:
                self._log.service('Failed to %s with angel: %s' % (cmd, e))
            raise ValueError('Failed to %s with angel: %s' % (cmd, e))
        else:
            if ok == CMD_OK:
                return response
            self._log.service('Failed to get OK from angel')
            raise ValueError('Failed to get OK from angel')
        self.client = None
    
    def get_service(self):
        self._log.service('Asking angel for service')
        return self._send_ok(CMD_GET_SERVICE)
    
    def started(self):
        self._log.service('Informing angel service ready')
        self._send_ok(CMD_STARTED)
        
    def set_service(self, serialised):
        self._log.service('Sending service to angel')
        self._send_ok(CMD_SET_SERVICE, serialised)
    
    def poll(self):
        self._send_ok(CMD_POLL)
        
    def log(self, lines):
        self._send_ok(CMD_LOG, lines)
    
    def __bool__(self):
        return self.client is not None
    __nonzero__ = __bool__



import os
from setuptools import setup

VERSION = "0.6.1"

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = "mara",
    version = VERSION,
    author = "Richard Terry",
    author_email = "code@radiac.net",
    description = ("A framework for network services, talkers and MUDs"),
    license = "BSD",
    keywords = "socket telnet",
    url = "http://richardterry.co.uk/projects/mara/",
    long_description=read('README.rst'),
    classifiers=[
        'Development Status :: 4 - Beta',
        #'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        "License :: OSI Approved :: BSD License",
        'Operating System :: OS Independent',
        "Topic :: Internet",
        "Topic :: Communications :: Chat",
        "Topic :: Games/Entertainment :: Multi-User Dungeons (MUD)",
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    install_requires=['six'],
    extras_require = {
        'dev':  ['nose'],
        'full': ['python-dateutil', 'pyyaml', 'bcrypt'],
    },
    test_suite = 'nose.collector',
    zip_safe=True,
    packages=['mara'],
    scripts=['bin/mara']
)



"""
Test the angel
"""
from __future__ import unicode_literals

import mara


###############################################################################
############################################################### Process
###############################################################################

if __name__ == '__main__':
    # test_angel.py run directly - has been called by the angel
    from examples import talker
    
    # Replace restart command so it doesn't require admin permission
    talker.commands.commands.unregister('restart')
    talker.commands.commands.register(
        'restart', mara.contrib.commands.cmd_restart,
    )
    
    talker.service.run()


###############################################################################
############################################################### Angel
###############################################################################

else:
    # test_angel.py imported - being called as a test
    
    import threading
    import shutil
    import os
    
    # Import lib and define constants here
    # It's a relative import, so can't import it if __main__ 
    from .lib import *
    ANGEL_SOCKET = os.path.join(EXAMPLES_DIR, 'test_angel.sock')
    TALKER_STORE = os.path.join(EXAMPLES_DIR, 'test_talker_store')
    
    
    class Angel(mara.angel.Angel):
        def collect_args(self):
            "Arguments for angel to start the service"
            return [
                # Service is in this file
                __file__,
                
                # Define settings - we have to do it here because we can't
                # access 
                '--root_path=%s' % EXAMPLES_DIR,
                '--angel_socket=%s' % ANGEL_SOCKET,
                '--store_path=%s' % TALKER_STORE,
                '--log=all' if DEBUG else '--no-log',
            ]
    
    
    class AngelThread(object):
        """
        Wrapper for managing an Angel in a thread
        
        Based on lib.TestService
        """
        def __init__(self):
            # Give service a thread using global tests.lib.thread_id
            self.thread_id = thread_counter.get()
            
            # Create angel
            self.angel = Angel(
                root_path=EXAMPLES_DIR,
                angel_socket=ANGEL_SOCKET,
                log='all' if DEBUG else False,
                settings_collect_args=False,
            )
        
            # Start service thread
            self._exception = None
            self.thread = threading.Thread(
                name="%s-%s" % (self.__class__.__name__, self.thread_id),
                target=self.run,
            )
            self.thread.daemon = True  # Kill with main process
            self.thread.start()
            
            # ++ wait for self.angel.active_process
            
            # Catch exception
            if self._exception:
                self.thread.join()
                raise RuntimeError("Thread %s failed" % self.thread.name)
        
        def run(self):
            """
            Run the angel. Call this as a thread target.
            """
            try:
                self.angel.run()
            except Exception as e:
                self._exception = e
                raise

        def stop(self):
            """
            Stop the angel
            """
            self.angel.stop()
            self.thread.join()


    from .test_talker import TalkerClient
    
    class AngelTest(TestCase):
        """
        Test the example talker using the angel
        """
        def remove_store(self):
            # Sanity check store so we don't delete the wrong thing
            if not TALKER_STORE or not CWD or not TALKER_STORE.startswith(CWD):
                raise ValueError('Invalid talker store')
            
            # Ensure test store doesn't exist
            if os.path.isdir(TALKER_STORE):
                shutil.rmtree(TALKER_STORE)
            
        def setUp(self):
            # Start service
            self.remove_store()
            self.angel = AngelThread()
            
        def tearDown(self):
            self.angel.stop()
            self.remove_store()
        
        def assertWho(self):
            self.ann.assertResponse(
                'who',
                hr('Currently online'),
                'ann\t0 seconds ago',
                'bob\t0 seconds ago',
                'cat\t0 seconds ago',
                hr(),
            )
            self.bob.assertNothing()
            self.cat.assertNothing()
            
        def test_multiple(self):
            "Test angel with multiple connections"
            # Create accounts
            self.ann = ann = TalkerClient(self)
            ann.create_account('ann', 'password1')
            ann.assertLine('Welcome, ann!', 'Nobody else is here.')
            
            self.bob = bob = TalkerClient(self)
            bob.create_account('bob', 'password2')
            bob.assertLine('Welcome, bob!', 'ann is here.')
            ann.assertLine('-- bob has connected --')
            
            self.cat = cat = TalkerClient(self)
            cat.create_account('cat', 'password3')
            cat.assertLine('Welcome, cat!', 'ann and bob are here.')
            ann.assertLine('-- cat has connected --')
            bob.assertLine('-- cat has connected --')
            
            # Confirm everything is working as expected
            self.assertWho()
            ann.assertResponse('say Hello', 'You say: Hello')
            bob.assertLine('ann says: Hello')
            cat.assertLine('ann says: Hello')
        
            # Issue the restart
            ann.assertResponse('restart', '-- Restarting server --')
            bob.assertLine('-- Restarting server --')
            cat.assertLine('-- Restarting server --')
            
            # Confirm everything is still working as expected
            self.assertWho()
            ann.assertResponse('say Hello', 'You say: Hello')
            bob.assertLine('ann says: Hello')
            cat.assertLine('ann says: Hello')



"""
Client container
"""
from __future__ import unicode_literals

from . import util


class ClientContainer(object):
    """
    Client container
    
    Holds a list of clients, allows filtering and bulk writes
    
    Containers are not directly managed across restarts; you should consider
    using storage.ClientContainer and storage.SessionClientContainer instead.
    """
    _clients = None
    
    def __init__(self, *args, **kwargs):
        """
        Initialise a container
        
        Designed to be mixed into other classes, so takes any arguments and
        ignores them.
        """
        self._clients = []
        super(ClientContainer, self).__init__(*args, **kwargs)
        
    clients = property(lambda self: self._clients)
    
    def add_client(self, client):
        self._clients.append(client)
        
    def remove_client(self, client):
        self._clients.remove(client.id)
    
    def write(self, clients, *data, **kwargs):
        """
        Send the provided lines to the given client, or list of clients
        """
        if not hasattr(clients, '__iter__'):
            clients = [clients]
        for client in clients:
            client.write(*data, **kwargs)
    
    def write_all(self, *data, **kwargs):
        """
        Send the provided lines to all active clients, as returned by
        filter_clients(**kwargs)
        
        Takes the same arguments as filter_clients()
        """
        # Get client list
        clients = self.filter_clients(**kwargs)
        
        # Write data to the clients
        for client in clients:
            client.write(*data)
    
    def filter_clients(self, filter_fn=None, exclude=None):
        """
        Get a list of clients, filtered by the global filter, a local filter
        and an exclude list
        """
        # Capture kwargs
        if exclude is None:
            exclude = []
        elif not hasattr(exclude, '__iter__'):
            exclude = [exclude]
            
        # Filter the client list using exclude list
        clients = [client for client in self.clients if client not in exclude]
        
        # Further filtering for this call
        if filter_fn:
            clients = filter_fn(self, clients)
        
        return clients
    
    msg_who = '%(others)s %(are)s %(where)s.'
    def get_who(self, exclude=None, where='here'):
        """
        Return a string of users who are in the container
        
        If client.user.name is available (ie contrib.user is being used) the
        names of the clients will be shown; otherwise it will be the number of
        clients (eg "5 are here")
        
        Fills out the string msg_who; the ``where`` argument can be overridden
        to change it from 'here', eg 'there', 'online' etc.
        """
        # Find others here
        other_clients = self.filter_clients(exclude=exclude)
        try:
            others = [client.user.name for client in other_clients]
        except AttributeError:
            others = ['%d' % len(other_clients)]
        else:
            if not others:
                if exclude:
                    others = ['Nobody else']
                else:
                    others = ['Nobody']
        
        # Welcome user
        return self.msg_who % {
            'others':   util.pretty_list(sorted(others)),
            'are':      'is' if len(others) == 1 else 'are',
            'where':    where,
        }
